(*******************************************************************)
(*  Avatar level 0++ driver.  Unit for providing a program         *)
(*  with proper Avatar levels 0, 0+, and 0++ emulations.           *)
(*  derived from my PAvatar/1 unit but excludes VT52, non-BBS,     *)
(*  and AVT/1 windowing, flow & sound functions.                   *)
(*  Copyright (c) 1991 Greg Smith                                  *)
(*-----------------------------------------------------------------*)
(*  Last Update:   March 15, 1992                                  *)
(*******************************************************************)
unit PAvt0; { AVT/0+ Driver (with ANSI-BBS fallback) }

{ DEFINE P_DEBUG}

interface

{$IFNDEF VER55}
{$X+,G-}
{$ENDIF}

{$S-,R-,V-,D-,L-}
{$IFDEF P_DEBUG}
{$D+,L+,S+,R+,G+}
{$ENDIF}

uses Dos, Crt;

type
  AvatarInterpProc = Procedure (ch:char); { interpreter procedure type }

var
  AVTInterp : AvatarInterpProc; { Next interp stage }

(*-- Declaration of Functions/Procedures useable by other programs --*)

procedure AvtTTY(ch:char);
{ Avatar TTY w/o AVT code processing }

procedure SetScreen(nc,nl:byte);
{ set the defualt screen size (columns and lines) }

function DESQview_Version: word;
{ return DESQview version number }

procedure DV_Pause;
{ give up DESQview time slice }

function ReadKey : char;
{ DESQview aware ReadKey replacement }


procedure ResetTerminal(nc,nl:byte);
{ Reset the Avatar terminal (columns and lines) }

procedure ANSI_Only;
{ Creates an ANSI terminal by putting into sleep & non-fallback mode }

function In_Command : boolean;
{ returns TRUE when the terminal is inside of an AVT command. }

procedure UpdateCursor(x,y:byte);
{ Re-Position the cursor within the terminal, x=0 means reset the position }

{-- public type declarations ---------------------------------------}
type  {-- The following is the list of Allowable terminal types ----}
  ReplyProc = Procedure (S:String);                    { Query Hook }
  WriteATHookT = Procedure (x,y,a:byte;ch:char); { for writing to screen }
  GotoXYHookT = Procedure (x,y:byte); { for moving cursor }
  ScrollHookT = Procedure (dir,x1,y1,x2,y2,n,a:byte); { Scrolling }
               { dir: 1=up,2=down,3=left,4=right. n=# 2 scroll,a=new attrib }
  FillHookT = Procedure (x1,y1,x2,y2,a:byte;ch:char); { filling an area }
  GetCharHookT = Procedure (x,y:byte;var a:byte;var c:char); { get scrn data }
  HighHookT = Procedure (x1,y1,x2,y2,a:byte); { for highlighting an area }
  AvtWindowRec = record { stores the current terminal setup }
                   cx,cy,  { Corresponds to WhereX/Y     }
                   width,  { pre-calculated Width        }
                   depth,  { the above, but depth        }
                   cattr : byte; { current attribute     }
                   insert : boolean; { Insert on?        }
                 end;


{-- public constants -----------------------------------------------}
const
  Fallback : boolean = True;
  QueryReply : string[80] =
               'AVT0, Avatar/0+ Copr. 1991 Greg Smith. Ver 1.17d'+^M;
  Dest_BS : boolean = True; { AVT/1 = destructive, AVT/0 = non-destructive }
  AvtWin : AvtWindowRec = (cx: 1; cy: 1; width: 80; depth: 25;
                           cattr: 3; insert: false);
  Sound_Stat : byte = $03; { Sound: bit0=bell }
  Esc = ^[;
  DLE = ^P;
  FS = #28;
  GS = #29;
  RS = #30;
  US = #31;
  SP = #32; { or ' ' }

{-- public variables -----------------------------------------------}
var
  Query_Hook : ReplyProc;    { processes Querey Replys             }
  WriteATh : WriteATHookT;   { \                                   }
  GotoXYh : GotoXYHookT;     {  \                                  }
  Scrollh : ScrollHookT;     {   > Hooks for screen manipulation   }
  FillAreah : FillHookT;     {  /                                  }
  GetATh : GetCharHookT;     { /                                   }
  HighAreah : HighHookT;     {/                                    }


implementation


{-- local variable -------------------------------------------------}
const
  Def_Attr = 3;                { defualt color in AVT/0 }

procedure SetScreen(nc,nl:byte);
begin
  with AvtWin do
   begin
     width := nc;
     depth := nl;
   end;
end;

Procedure SetXY(x,y:byte);
begin
  with AvtWin do
   begin
     if x > Width then x := Width;
     if y > Depth then y := Depth;
     if x = 0 then x := 1;
     if y = 0 then y := 1;
     cx := x;
     cy := y;
     GotoXYh(x,y);
   end;
end;

Function WhereX: byte;
begin
  WhereX := AvtWin.cx;
end;

Function WhereY: byte;
begin
  WhereY := AvtWin.cy;
end;

Procedure HighArea(x1,y1,x2,y2,a:byte);
begin
  with AvtWin do
   HighAreah(x1,y1,x2,y2,a);
end;

Procedure Scroll(dir,x1,y1,x2,y2,n:byte);
begin
  with AvtWin do
   begin
     if n = 0 then
      begin
        if dir < 3 then n := succ(y2-y1)
         else n := succ(x2-x1);
      end;
     Scrollh(dir,x1,y1,x2,y2,n,cattr);
   end;
end;

Procedure FillArea(x1,y1,x2,y2,a:byte;c:char);
begin
  with AvtWin do
   begin
     if x2 > Width then x2 := Width;
     if y2 > Depth then y2 := Depth;
     FillAreah(x1,y1,x2,y2,a,c);
   end;
end;

Procedure DDelEOL;
begin
  with AvtWin do
   FillArea(cx,cy,Width,cy,cattr,' ');
end;

Procedure DDelChar;
begin
  with AvtWin do
   Scroll(3,cx,cy,Width,cy,1); { normal }
end;

Procedure DDelLine; { delete line }
begin
  with AvtWin do
   Scroll(1,1,succ(cy),Width,Depth,1) { delete line }
end; { ddelline }

Procedure DInsertLine; { insert line }
begin
  with AvtWin do
   Scroll(2,1,cy,Width,pred(Depth),1)  { insert line }
end; { dinsertline }

Procedure DHighEOL(a : byte); { Highlight EOL }
begin
  with AvtWin do
   HighArea(cx,cy,Width,cy,a); { normal }
end; { DHighEOL }

Procedure DHighBOL(a : byte); { Highlight BOL }
begin
  with AvtWin do
   HighArea(1,cy,cx,cy,a); { normal }
end; { DHighBOL }

Procedure HighlightIt(a : byte);
begin
  with AvtWin do
   HighArea(1,1,Width,Depth,a);
end; { HighlightIt }

Procedure Clear(a:byte;c:char); { Clear the Screen }
begin
  with AvtWin do
   FillArea(1,1,Width,Depth,a,c);
end;

Procedure WriteAt(x,y,a:byte;c:char);
begin
  with AvtWin do
   WriteATh(x,y,a,c);
end;

{$F+ compatible with TP 5.5 }
Procedure NullReply(s:string); begin end;
Procedure NullWriteAT(x,y,a:byte;c:char); begin end;
Procedure NullGotoXY(x,y:byte); begin end;
Procedure NullScroll(d,x,y,x1,y1,n,a:byte); begin end;
Procedure NullFill(x,y,x1,y1,a:byte;c:char); begin end;
Procedure NullGetChar(x,y:byte;var a:byte;var c:char); begin end;
Procedure NullHigh(x,y,x1,y1,a:byte); begin end;
{$F-}

Procedure Nullify_Hooks;
begin
  Query_Hook  := NullReply;
  WriteATh    := NullWriteAT;
  GotoXYh     := NullGotoXY;
  Scrollh     := NullScroll;
  FillAreah   := NullFill;
  GetATh      := NullGetChar;
  HighAreah   := NullHigh;
end;

(* ---  More locals  ----- *)

type
  Arr255Type = Array[1..255] of char;

const
  AvtIncome   : byte = 0; { # of parameters expected to come }
  AvtInString : byte = 0; { nesting level of ^V^Y command }
  PrmArr : Array[1..10] of byte = (0,0,0,0,0,0,0,0,0,0); { stores params }
  MaxStrNest = 20; { Maximum nesting level of ^V^Y commands }

var
  AvtStrPtr : Array[1..MaxStrNest] of ^Arr255Type; { strs storing ^V^Y cmds }
  AvtInterp_Ptr : pointer absolute AvtInterp;

{-- Forward Interp References -------------------------------------}

{$F+}
procedure AVT0_Waiting(ch:char); forward;
procedure AVT0_Command(ch:char); forward;
procedure AVT0_Param(ch:char); forward;
procedure AVT0_String(ch:char); forward;
{$F-}

{-- ANSI & DV Keyboard Include Files ------------------------------}

{$I ANSIBBS.INC  }
{$I AVTDESQV.INC }

{-- Inner workings of terminal ------------------------------------}

function In_Command : boolean; { Is the terminal inside a command? }
begin
  In_Command := (Word(AvtInterp_Ptr) <> Word(@AVT0_Waiting)) and
                (Word(AvtInterp_Ptr) <> Word(@ANSI_Waiting)) or
                (AvtInString > 0);
end;

procedure UpdateCursor(x,y:byte);
begin
 with AvtWin do
  if (x = 0) then
   SetXY(cx,cy) { restore position }
  else
   SetXY(x,y);          { give it a new position }
end;

procedure ResetTerminal(nc,nl:byte);
begin
  SetScreen(nc,nl);
  with AvtWin do
   begin
     cattr := Def_Attr;
     insert := false;
   end;
  ANSI_Reset;
  AvtInCome := 0;
  AvtInString := 0;
  AvtInterp := AVT0_Waiting; { waiting }
  Fallback := True;
  Dest_BS := False;
end; { ResetTerminal }

procedure AvtTTY(ch:char);
var
  oc,n : byte;
  WhX,WhY : byte;
begin
  with AvtWin do
   begin
     WhX := cx;
     WhY := cy;
     if ch in [^G,^H,^I,^J,^L,^M] then
      case ch of
         ^M : begin { CR }
                SetXY(1,WhY); { left of screen }
              end; { ^M }
         ^J : begin { LF }
                if WhY >= Depth then
                 Scroll(1,1,1,Width,Depth,1)
                else
                 if Insert then DInsertLine; { Insert a line if active }
                SetXY(WhX,succ(WhY)); { down one line }
              end; { ^J }
         ^H : begin { BS }
                if Dest_BS and (WhX > 1) then
                 begin
                   WriteAth(pred(WhX),WhY,cattr,' '); { destructive? }
                   if Insert then { scroll line left if insert is on }
                    Scroll(3,pred(WhX),WhY,Width,WhY,1);
                 end;
                SetXY(pred(WhX),WhY); { normal }
              end; { ^H }
         ^I : begin { Tab 8 }
                if Insert then
                 begin
                   if WhX < (Width-8) then
                    Scroll(4,WhX,WhY,Width-8,WhY,8);
                 end;
                SetXY(WhX+8,WhY); { normal }
              end; { ^I }
         ^G : begin { beep }
                if ((Sound_Stat and 1) = 1) then  { bell on }
                 begin
                   Sound(262);
                   delay(100); { 1/10th second low C for a beep }
                   NoSound;
                 end;
              end; { ^G }
         ^L : begin { cls, already caught by AVT/0, but not ANSI }
                Clear(7,' '); { defualt is 7 in ANSI }
                SetXY(1,1);
              end;
      end { case }
     else { if ch in ... }
      begin
        if Insert then
         begin
           if WhX < Width then
            Scroll(4,WhX,WhY,Width,WhY,1);
         end; { insert }
        WriteAT(WhX,WhY,cattr,ch); { write the char }
         if WhX < Width then
          SetXY(succ(WhX),WhY)
         else
          begin
            if WhY >= Depth then
             Scroll(1,1,1,Width,Depth,1); { scroll the screen }
            SetXY(1,succ(WhY));
          end; { else }
      end; { if ch in ... else (write the char) }
   end; { With AvtWin }
end; { AvtTTY }

(*************** Query Handler *******************************)

Procedure Process_Query(ch : char);
begin
  with AvtWin do
   begin
     case chr(ord(ch) and $3F) of
       ^A,^B : begin
                 Query_Hook(^V+^A+chr(cattr));
               end;
       ^H : begin
              Query_Hook(^V+^H+chr(cy)+chr(cx));
            end;
       ^Q : begin  { normal query }
              Query_Hook(QueryReply);
            end;
     end; { case } { all queries which could be part of AVT/0+ }
   end; { with }
end; { Process_Query }

(*************** Avatar level 0+ Interpreter ******************)

procedure AVT0_Waiting(ch:char); {far;}
begin
  if ch > #27 then
   AvtTTY(ch) { TTY char }
  else
   begin
     case ch of
       ^V : AVTInterp := AVT0_Command;
       ^L : begin
              Clear(Def_Attr,' '); { ClrScr }
              SetXY(1,1);
              AvtWin.Insert := False; { Turn Off Insert }
              exit;
            end;
       ^Y : begin
              AvtInterp := AVT0_Param;
              AvtInCome := 2; { expect 2 more params. }
            end;
      ESC : begin
              if (AvtInString = 0) and Fallback then
               AvtInterp := ANSI_Bracket
              else AvtTTY(ESC);
            end;
      else { case }
       begin
         AvtTTY(ch);
         exit;
       end; { case else }
     end; { case }
     PrmArr[1] := ord(ch);  { put command in parameter array }
   end; { ch <= #27 }
end;

procedure AVT0_Command(ch:char); {far;}
begin                                    { primary interpretation phase }
  AvtInterp := AVT0_Waiting; { waiting }
  PrmArr[2] := ord(ch); { store the command }
  if ch <> ^Y then AvtWin.Insert := False; { no insert }
  case ch of   { order of frequency. }
    ^A : begin  { set attribute }
           AvtInCome := 1;
           AvtInterp := AVT0_Param;
         end;
    ^H : begin  { set cursor position }
           AvtInCome := 2;
           AvtInterp := AVT0_Param;
         end;
    ^B : begin  { turn blink on }
           with AvtWin do
            cattr := cattr or $80;
         end;
    ^E : begin  { move cursor one column left }
           with AvtWin do
            SetXY(pred(cx),cy);
         end;
    ^F : begin  { Move cursor one column right }
           with AvtWin do
            SetXY(succ(cx),cy);
         end;
    ^C : begin  { move cursor one line up }
           with AvtWin do
            SetXY(cx,pred(cy));
         end;
    ^D : begin  { move cursor one line down }
           with AvtWin do
            SetXY(cx,succ(cy));
         end;
    ^J,^K : begin  { scroll up/down }
           AvtInCome := 5;
           AvtInterp := AVT0_Param;
           if ch = ^J then PrmArr[7] := 1 else PrmArr[7] := 2;
         end;
    ^L : begin  { clear an area }
           AvtInCome := 3;
           AvtInterp := AVT0_Param;
         end;
    ^I : begin  { turn insert on }
           AvtWin.Insert := True;
         end;
    ^M : begin  { initalize an area }
           AvtInCome := 4;
           AvtInterp := AVT0_Param;
         end;
    ^G : begin  { clear to EOL }
           DDelEOL; { easy huh? }
         end;
    ^N : begin  { delete character }
           DDelChar; { another easy one! }
         end;
   '-' : begin  { delete line }
           DDelLine; { easy again?! }
         end;
   '+' : begin { insert line }
           DInsertLine;
         end;
   '>','<' : begin { scroll right/left }
           AvtInCome := 5;
           AvtInterp := AVT0_Param;
           if ch = '<' then PrmArr[7] := 3 else PrmArr[7] := 4;
         end;
    ^R : begin { reset the driver }
           AvtInCome := 0;
           with AvtWin do
            ResetTerminal(width,depth);
         end;
   '!' : begin { Poke char on screen }
           AvtInCome := 4;
           AvtInterp := AVT0_Param;
         end;
   '?' : begin { Peek at char on screen }
           AvtInCome := 2;
           AvtInterp := AVT0_Param;
         end;
   ^Q,^T,         { Query, Highlight position }
   '0','1': begin { Highlight EOL/BOL }
              AvtInCome := 1;
              AvtInterp := AVT0_Param;
            end;
    ^Y : begin { repeat AVT pattern }
           AvtInCome := 255; { undefined as of yet }
           AvtInterp := AVT0_Param;
         end;
  end; { case ch of }
end; { AVT0_Command }

Procedure AVT0_Param(ch:char); {far;}
var
  b : Array[1..6] of byte;
  i1,i2 : integer;
  c1 : char absolute b; { for referencing purposes }
begin
  if PrmArr[1] = 22 then { if ^V then }
  case chr(PrmArr[2]) of
    ^A : begin { Set Attribute }
           AvtWin.cattr := (ord(ch) AND $7F);
           AvtInCome := 0;
         end;
    ^H : begin
           if AvtInCome = 1 then
            SetXY(ord(ch),PrmArr[3])
            else
             PrmArr[5-AvtInCome] := ord(ch);
           dec(AvtInCome);
         end;
   ^J,^K,'<','>' : begin { scroll up/down/left/right }
           if AvtInCome > 1 then
            begin
              PrmArr[8-AvtInCome] := ord(ch);
              dec(AvtInCome);
            end
           else
            begin                     { dir   ,  x1   ,    y1   }
              Scroll(PrmArr[7],PrmArr[5],PrmArr[4],
               ord(ch),PrmArr[6],PrmArr[3]);  { x2, y2, #lines }
              AvtInCome := 0;
            end;
         end; { ^J/^K }
    ^L : begin { clear area }
           if AvtInCome > 1 then
            begin
              PrmArr[6-AvtInCome] := ord(ch);
              dec(AvtInCome);
            end
           else
            begin
              with AvtWin do
               begin
                 FillArea(cx,cy,pred(cx+ord(ch)),
                          pred(cy+PrmArr[4]),PrmArr[3],' ');
                 cattr := PrmArr[3];
               end;
              AvtInCome := 0;
            end;
         end;
    ^M : begin { init area }
           if AvtInCome > 1 then
            begin
              PrmArr[7-AvtInCome] := ord(ch);
              dec(AvtInCome);
            end
           else
            begin
              with AvtWin do
               begin
                 FillArea(cx,cy,pred(cx+ord(ch)),
                          pred(cy+PrmArr[5]),PrmArr[3],
                          chr(PrmArr[4]));
                 cattr := PrmArr[3];
               end;
              AvtInCome := 0;
            end;
         end;
    ^Q : begin { Query the driver }
           Process_Query(ch);
           AvtInCome := 0;
         end;
    ^T : begin { highlight cursor position }
           with AvtWin do
            HighArea(cx,cy,cx,cy,ord(ch));
           AvtInCome := 0;
         end;
   '!' : begin { Poke character on physical screen }
           if AvtInCome > 1 then
            begin
              PrmArr[7-AvtInCome] := ord(ch);
              dec(AvtInCome);
            end
           else
            begin
              WriteATh(ord(ch),PrmArr[5],PrmArr[4],
                       chr(PrmArr[3]));
              AvtInCome := 0;
            end;
         end; { ! }
   '0' : begin { Highlight EOL }
           DHighEOL(ord(ch));
           AvtInCome := 0;
         end; { 0 }
   '1' : begin { Highlight BOL }
           DHighEOL(ord(ch));
           AvtInCome := 0;
         end; { 1 }
   '?' : begin { peek at physical screen }
           if AvtInCome > 1 then
            begin
              PrmArr[3] := ord(ch);
              Dec(AvtInCome);
            end
           else
            begin
              GetATh(ord(ch),PrmArr[3],b[2],c1);
              Query_Hook(^V+'!'+chr(PrmArr[3])+ch+chr(b[2])+c1);
              AvtInCome := 0;
            end;
         end; { ? }
    ^Y : begin { get the defined params for ^V^Y }
           PrmArr[3] := ord(ch); { # of bytes in pattern }
           AvtInCome := ord(ch);
           inc(AvtInString); { add 1 to str nesting level }
           if AvtInString > MaxStrNest then
            begin
              if AvtInCome > 0 then
               begin
                 dec(AvtInCome);
                 AVTInterp(ch); { parse it, but not repetitous }
               end
              else
               AvtInterp := AVT0_Waiting; { exit failed-rep pattern }
            end   { too many strings }
           else
            if ord(ch) > 0 then
             begin
               AvtInterp := AVT0_String;
               PrmArr[9] := ord(ch);
               GetMem(AvtStrPtr[AvtInString],PrmArr[9]);
               if AvtStrPtr[AvtInString] = nil then
                begin
                  Writeln('AVATAR1: GetMem Failure.');
                  AvtInterp := AVT0_Waiting; { abort sequence }
                  AvtInCome := 0;
                end;
             end; { length > 0 }
           exit;
         end; { ^Y }
  end { GetParm case }
  else { if PrmArr[1] = 22 }
   if PrmArr[1] = 25 then
    begin
      if AvtInCome = 1 then
       for PrmArr[3] := ord(ch) downto 1 do
        AvtTTY(chr(PrmArr[2]))
       else
        PrmArr[2] := ord(ch);
      dec(AvtInCome);
    end; { if = 25 and if=22-else}
  if AvtInCome = 0 then AvtInterp := AVT0_Waiting;
end; { AVT0_Param }

Procedure AVT0_String(ch:char); {far;}
var
  b : Array[1..4] of byte;
begin
  if AvtInCome >= 1 then
   begin
     dec(AvtInCome);     { Add the char to the string }
     AvtStrPtr[AvtInString]^[PrmArr[3]-AvtInCome] := ch;
   end
  else
   begin
     AvtInterp := AVT0_Waiting;
     b[3] := PrmArr[3];
     b[4] := AvtInString;
     for b[1] := 1 to ord(ch) do { repeat # times }
      for b[2] := 1 to b[3] do { each char in string }
       AVTInterp(AvtStrPtr[b[4]]^[b[2]]);
     FreeMem(AvtStrPtr[AvtInString],PrmArr[9]); { Free it up }
     dec(AvtInString);
     AvtInterp := AVT0_Waiting;
   end;
end; { AVT0_String }

(************* Special Effects Procedures *********************)

procedure ANSI_Only; { Disable AVT/0++ commands }
begin
  fallback := false;
  AvtInterp := ANSI_Waiting;
end;

(***************** Unit Initialization ***********************)

{$S+} Procedure InitUnit; {$S-}
var
  i : integer;
begin
  DESQview_Init;
  Nullify_Hooks;
  ResetTerminal(80,25);
end;

Begin
  InitUnit;
End.
